@@ -2,6 +2,7 @@
 #define G_LOG_DOMAIN "rauc-nbd"
 
 #include <unistd.h>
+#include <grp.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
@@ -86,8 +87,9 @@ void r_nbd_free_server(RaucNBDServer *nbd_srv)
 	g_free(nbd_srv->tls_key);
 	g_free(nbd_srv->tls_ca);
 	g_strfreev(nbd_srv->headers);
-	g_strfreev(nbd_srv->info_headers);
+	g_clear_pointer(&nbd_srv->info_headers, g_ptr_array_unref);
 	g_free(nbd_srv->effective_url);
+	g_free(nbd_srv->etag);
 	g_free(nbd_srv);
 }
 
@@ -140,7 +142,7 @@ out:
 	return nl;
 }
 
-gboolean r_nbd_setup_device(RaucNBDDevice *nbd_dev, GError **error)
+gboolean r_nbd_setup_device(RaucNBDDevice *nbd_dev, int *devicefd, GError **error)
 {
 	GError *ierror = NULL;
 	gboolean res = FALSE;
@@ -149,8 +151,11 @@ gboolean r_nbd_setup_device(RaucNBDDevice *nbd_dev, GError **error)
 	struct nl_msg *msg = NULL;
 	struct nlattr *attr_sockets = NULL;
 	struct nlattr *attr_item = NULL;
+	g_autofree gchar *device_path = NULL;
+	g_auto(filedesc) idevicefd = -1;
 
 	g_return_val_if_fail(nbd_dev != NULL, FALSE);
+	g_return_val_if_fail(devicefd != NULL && *devicefd == -1, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
 	g_assert(nbd_dev->data_size % 4096 == 0);
@@ -175,9 +180,6 @@ gboolean r_nbd_setup_device(RaucNBDDevice *nbd_dev, GError **error)
 	NLA_PUT_U64(msg, NBD_ATTR_SIZE_BYTES, nbd_dev->data_size);
 	NLA_PUT_U64(msg, NBD_ATTR_BLOCK_SIZE_BYTES, 4096);
 	NLA_PUT_U64(msg, NBD_ATTR_SERVER_FLAGS, 0);
-	NLA_PUT_U64(msg, NBD_ATTR_CLIENT_FLAGS,
-			NBD_CFLAG_DISCONNECT_ON_CLOSE
-			);
 	NLA_PUT_U64(msg, NBD_ATTR_TIMEOUT, 300);
 
 	attr_sockets = nla_nest_start(msg, NBD_ATTR_SOCKETS);
@@ -199,11 +201,40 @@ gboolean r_nbd_setup_device(RaucNBDDevice *nbd_dev, GError **error)
 	if (!nbd_dev->index_valid)
 		g_error("failed to create nbd device");
 
-	nbd_dev->dev = g_strdup_printf("/dev/nbd%"G_GUINT32_FORMAT, nbd_dev->index);
+	device_path = g_strdup_printf("/dev/nbd%"G_GUINT32_FORMAT, nbd_dev->index);
+
+	idevicefd = g_open(device_path, O_RDONLY | O_CLOEXEC);
+	if (idevicefd < 0) {
+		int err = errno;
+		g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(err), "failed to open %s: %s", device_path, g_strerror(err));
+		res = FALSE;
+		goto out;
+	}
+
+	msg = nlmsg_alloc();
+	if (!msg)
+		g_error("failed to allocate netlink message");
+
+	if (!genlmsg_put(msg, NL_AUTO_PORT, NL_AUTO_SEQ, driver_id, 0, 0, NBD_CMD_RECONFIGURE, 0))
+		g_error("failed to add generic netlink headers to message");
+
+	NLA_PUT_U32(msg, NBD_ATTR_INDEX, nbd_dev->index);
+	NLA_PUT_U64(msg, NBD_ATTR_CLIENT_FLAGS, NBD_CFLAG_DISCONNECT_ON_CLOSE);
+
+	nl_socket_modify_cb(nl, NL_CB_VALID, NL_CB_CUSTOM, NULL, NULL);
+	if (nl_send_sync(nl, msg) < 0) {
+		res = FALSE;
+		g_set_error(error, G_IO_ERROR, G_IO_ERROR_FAILED, "netlink send_sync failed");
+		goto out;
+	}
 
 	g_message("setup done for %s", nbd_dev->dev);
 
 	res = TRUE;
+	nbd_dev->dev = g_steal_pointer(&device_path);
+	*devicefd = idevicefd;
+	idevicefd = -1;
+	(void)idevicefd; /* ignore dead store, replace with g_steal_fd when we have glib 2.70 */
 	goto out;
 
 	/* This label is used by the NLA_PUT macros. */
@@ -318,8 +349,17 @@ struct RaucNBDTransfer {
 	guint64 content_size;
 	guint64 current_time; /* date header from server */
 	guint64 modified_time; /* last-modified header from server */
+	gchar *etag;
 };
 
+static void free_transfer(struct RaucNBDTransfer *xfer)
+{
+	g_clear_pointer(&xfer->buffer, g_free);
+	g_clear_pointer(&xfer->etag, g_free);
+
+	g_free(xfer);
+}
+
 static size_t write_cb(char *ptr, size_t size, size_t nmemb, void *userdata)
 {
 	struct RaucNBDTransfer *xfer = userdata;
@@ -398,6 +438,10 @@ static size_t header_cb(char *buffer, size_t size, size_t nitems, void *userdata
 			xfer->modified_time = date;
 			g_message("nbd server received HTTP file date %"G_GUINT64_FORMAT, xfer->modified_time);
 		}
+	} else if (g_str_equal(h_name, "etag")) {
+		r_replace_strdup(&xfer->etag, h_pair[1]);
+		g_autofree gchar *escaped = g_strescape(h_pair[1], NULL);
+		g_message("nbd server received HTTP ETag: \"%s\"", escaped);
 	}
 
 	return nitems;
@@ -421,8 +465,16 @@ static void prepare_curl(struct RaucNBDTransfer *xfer)
 	code |= curl_easy_setopt(xfer->easy, CURLOPT_URL, xfer->ctx->url);
 	if (xfer->ctx->tls_cert)
 		code |= curl_easy_setopt(xfer->easy, CURLOPT_SSLCERT, xfer->ctx->tls_cert);
-	if (xfer->ctx->tls_key)
+	if (xfer->ctx->tls_key) {
+		if (g_str_has_prefix(xfer->ctx->tls_key, "pkcs11:")) {
+#if ENABLE_OPENSSL_PKCS11_ENGINE
+			code |= curl_easy_setopt(xfer->easy, CURLOPT_SSLKEYTYPE, "ENG");
+#else
+			code |= curl_easy_setopt(xfer->easy, CURLOPT_SSLKEYTYPE, "PROV");
+#endif
+		}
 		code |= curl_easy_setopt(xfer->easy, CURLOPT_SSLKEY, xfer->ctx->tls_key);
+	}
 	if (xfer->ctx->tls_ca) {
 		code |= curl_easy_setopt(xfer->easy, CURLOPT_CAINFO, xfer->ctx->tls_ca);
 		code |= curl_easy_setopt(xfer->easy, CURLOPT_CAPATH, NULL);
@@ -527,7 +579,7 @@ static void start_read(struct RaucNBDContext *ctx, struct RaucNBDTransfer *xfer)
 		g_error("unexpected error from curl_multi_add_handle in %s", G_STRFUNC);
 }
 
-/* Appends Gstrv elements to curl_slist (strings are copied).
+/* Appends GStrv elements to curl_slist (strings are copied).
  * If curl_slist does not exist yet (NULL passed), it will be created.
  * The created list needs to be freed (after usage) by the caller with
  * curl_slist_free_all(). */
@@ -637,7 +689,7 @@ static void start_request(struct RaucNBDContext *ctx, struct RaucNBDTransfer *xf
 		case NBD_CMD_DISC: {
 			g_message("nbd server received disconnect request");
 			ctx->done = TRUE;
-			g_free(xfer); /* not queued via curl_multi_add_handle */
+			free_transfer(xfer); /* not queued via curl_multi_add_handle */
 			break;
 		}
 		case RAUC_NBD_CMD_CONFIGURE: {
@@ -732,7 +784,7 @@ static gboolean finish_configure(struct RaucNBDContext *ctx, struct RaucNBDTrans
 				error = g_strdup_printf("not modified");
 				break;
 			default:
-				error = g_strdup_printf("unexpected HTTP error code %ld", response_code);
+				error = g_strdup_printf("unexpected HTTP response code %ld", response_code);
 		}
 		g_variant_dict_insert(&dict, "error", "s", error);
 		g_variant_dict_insert(&dict, "error-http-code", "u", (guint32)response_code);
@@ -787,6 +839,8 @@ reply:
 		g_variant_dict_insert(&dict, "current-time", "t", xfer->current_time);
 	if (xfer->modified_time)
 		g_variant_dict_insert(&dict, "modified-time", "t", xfer->modified_time);
+	if (xfer->etag)
+		g_variant_dict_insert(&dict, "etag", "s", xfer->etag);
 
 	v = g_variant_dict_end(&dict);
 	reply_size = g_variant_get_size(v);
@@ -951,12 +1005,12 @@ gboolean r_nbd_run_server(gint sock, GError **error)
 						error,
 						R_NBD_ERROR, R_NBD_ERROR_SHUTDOWN,
 						"finish_request failed, shutting down");
-				g_free(xfer);
+				free_transfer(xfer);
 				goto out;
 			}
 
 			if (xfer->done) {
-				g_free(xfer);
+				free_transfer(xfer);
 			} else {
 				/* retry */
 				sleep(1);
@@ -1032,7 +1086,8 @@ static gboolean nbd_configure(RaucNBDServer *nbd_srv, GError **error)
 	if (nbd_srv->headers)
 		g_variant_dict_insert(&dict, "headers", "^as", nbd_srv->headers);
 	if (nbd_srv->info_headers)
-		g_variant_dict_insert(&dict, "info-headers", "^as", nbd_srv->info_headers);
+		g_variant_dict_insert(&dict, "info-headers", "@as",
+				g_variant_new_strv((const gchar **)nbd_srv->info_headers->pdata, nbd_srv->info_headers->len));
 	v = g_variant_dict_end(&dict);
 	{
 		g_autofree gchar *tmp = g_variant_print(v, TRUE);
@@ -1079,7 +1134,7 @@ static gboolean nbd_configure(RaucNBDServer *nbd_srv, GError **error)
 	if (reply_error) {
 		guint32 http_code = 0;
 		g_variant_dict_lookup(&dict, "error-http-code", "u", &http_code);
-		g_message("received http error %"G_GUINT32_FORMAT, http_code);
+		g_message("received HTTP response code %"G_GUINT32_FORMAT, http_code);
 		if (http_code == 204) {
 			g_set_error(
 					error,
@@ -1131,20 +1186,36 @@ static gboolean nbd_configure(RaucNBDServer *nbd_srv, GError **error)
 		g_autofree gchar *formatted_date = g_date_time_format(datetime, "%Y-%m-%d %H:%M:%S");
 		g_message("received HTTP server info: modified time %s (%"G_GUINT64_FORMAT ")", formatted_date, nbd_srv->modified_time);
 	}
+	g_variant_dict_lookup(&dict, "etag", "s", &nbd_srv->etag);
 
 	return TRUE;
 }
 
-struct child_setup_args {
+typedef struct {
 	uid_t uid;
 	gid_t gid;
-};
+	gid_t *groups;
+	int ngroups;
+} child_setup_args;
+
+static void clear_child_setup_args(child_setup_args *child_args)
+{
+	g_clear_pointer(&child_args->groups, g_free);
+}
+G_DEFINE_AUTO_CLEANUP_CLEAR_FUNC(child_setup_args, clear_child_setup_args);
 
 static void nbd_server_child_setup(gpointer user_data)
 {
 	/* see signal-safety(7) for functions which can be used here */
-	struct child_setup_args *args = user_data;
+	child_setup_args *args = user_data;
 
+	if (args->groups) {
+		if (setgroups(args->ngroups, args->groups) == -1) {
+			const char *msg = "setgroups failed\n";
+			write(STDOUT_FILENO, msg, strlen(msg));
+			exit(1);
+		}
+	}
 	if (args->gid) {
 		if (setgid(args->gid) == -1) {
 			const char *msg = "setgid failed\n";
@@ -1161,14 +1232,16 @@ static void nbd_server_child_setup(gpointer user_data)
 	}
 }
 
-static gboolean nbd_server_child_prepare(struct child_setup_args *args, GError **error)
+static gboolean nbd_server_child_prepare(child_setup_args *args, GError **error)
 {
 	const gchar *user = NULL;
 	struct passwd passwd = {0};
 	struct passwd *result = NULL;
 	g_autofree gchar *buf = NULL;
+	g_autofree gid_t *groups = NULL;
 	long bufsize;
 	int err;
+	int ngroups;
 
 	g_return_val_if_fail(args != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
@@ -1204,8 +1277,27 @@ static gboolean nbd_server_child_prepare(struct child_setup_args *args, GError *
 		}
 	}
 
+	ngroups = 0;
+	/* getgrouplist always fail (returns -1), but still sets ngroups to the
+	 * number of wanted groups. */
+	getgrouplist(user, result->pw_gid, NULL, &ngroups);
+	groups = g_malloc0(ngroups * sizeof(*groups));
+	args->ngroups = ngroups;
+
+	err = getgrouplist(user, result->pw_gid, groups, &args->ngroups);
+	/* Something very weird happened if the number of groups now is different
+	 * than the expected number from before. */
+	if (err < 0 || args->ngroups != ngroups) {
+		g_set_error(
+				error,
+				R_NBD_ERROR, R_NBD_ERROR_STARTUP,
+				"cannot get groups for user %s", user);
+		return FALSE;
+	}
+
 	args->uid = result->pw_uid;
 	args->gid = result->pw_gid;
+	args->groups = g_steal_pointer(&groups);
 
 	return TRUE;
 }
@@ -1232,7 +1324,7 @@ gboolean r_nbd_start_server(RaucNBDServer *nbd_srv, GError **error)
 	}
 
 	if (1) { /* subprocess */
-		struct child_setup_args child_args = {0};
+		g_auto(child_setup_args) child_args = {0};
 		g_autofree gchar *executable = NULL;
 		g_autoptr(GSubprocessLauncher) launcher = NULL;
 		g_autoptr(GPtrArray) args = g_ptr_array_new_full(3, g_free);
@@ -1257,6 +1349,7 @@ gboolean r_nbd_start_server(RaucNBDServer *nbd_srv, GError **error)
 		g_subprocess_launcher_set_child_setup(launcher, nbd_server_child_setup, &child_args, NULL);
 		g_subprocess_launcher_setenv(launcher, "RAUC_NBD_SERVER", "", TRUE);
 		g_subprocess_launcher_take_fd(launcher, sockets[0], RAUC_SOCKET_FD);
+		sockets[0] = -1; /* GSubprocessLauncher takes ownership */
 
 		nbd_srv->sproc = r_subprocess_launcher_spawnv(launcher, args, &ierror);
 		if (nbd_srv->sproc == NULL) {
@@ -1273,8 +1366,6 @@ gboolean r_nbd_start_server(RaucNBDServer *nbd_srv, GError **error)
 		g_thread_new("nbd", nbd_server_thread, sockp);
 	}
 
-	sockets[0] = -1; /* GSubprocess takes ownership */
-
 	nbd_srv->sock = sockets[1];
 	sockets[1] = -1; /* RaucNBDServer takes ownership */
 
@@ -1346,7 +1437,6 @@ gboolean r_nbd_read(gint sock, guint8 *data, size_t size, off_t offset, GError *
 {
 	struct nbd_request request = {0};
 	struct nbd_reply reply = {0};
-	gboolean res = FALSE;
 
 	g_return_val_if_fail(sock >= 0, FALSE);
 	g_return_val_if_fail(data, FALSE);
@@ -1373,14 +1463,11 @@ gboolean r_nbd_read(gint sock, guint8 *data, size_t size, off_t offset, GError *
 				error,
 				R_NBD_ERROR, R_NBD_ERROR_READ,
 				"failed to read data from remote server");
-		goto out;
+		return FALSE;
 	}
 
 	if (!r_read_exact(sock, data, size, NULL))
 		g_error("failed to receive nbd read reply body");
 
-	res = TRUE;
-
-out:
-	return res;
+	return TRUE;
 }
